/**
 * This code accompanies the paper:
 *
 * "Dense non-rigid shape correspondence using random forests"
 * E.Rodola, S.Rota Bulo, T.Windheuser, M.Vestner, D.Cremers.
 * Proc. CVPR 2014
 *
 * Please cite the paper above if you use the code.
 */
 
#ifndef SHAPEFCLASSIFIER_H_
#define SHAPEFCLASSIFIER_H_

#include "RandomForestCore.h"

#include <vector>
#include <limits>
#include <ctime>
#include <cmath>
#include <map>
#include <boost/cstdint.hpp>

using namespace vision;
using boost::uint32_t;

namespace vision {

struct WaveDescriptorParams
{
      WaveDescriptorParams() : energy_level(0), nEigenVectors(0) {}
      int energy_level;
      uint32_t nEigenVectors;
};

struct ShapeSplitData {

      ShapeSplitData() :
         type(0), thres(0) {
      }

      WaveDescriptorParams measurements[2];

      short type;
      double thres;
};

class ShapePrediction {
   public:
      ShapePrediction() {
      }

      void reset() {
         prob.clear();
      }

      void add(uint32_t label, float x) {
         assert(prob.find(label)!=prob.end()||prob[label]==0); // TODO remove
         prob[label]+=x;
      }

      void add(const ShapePrediction &p) {
         for (std::map<uint32_t,float>::const_iterator it=p.prob.begin(); it!=p.prob.end(); ++it) {
            assert(prob.find(it->first)!=prob.end()||prob[it->first]==0); // TODO remove
            prob[it->first]+=it->second;
         }
      }

      void normalize() {
         float sum=0;
         std::map<uint32_t,float>::iterator it;
         for (it=prob.begin(); it!=prob.end(); ++it)
            sum+=it->second;
         for (it=prob.begin(); it!=prob.end(); ++it)
            it->second/=sum;
      }

      uint32_t bestLabel() {
         uint32_t bestLabelIdx=0;
         float maxProb=0;
         if (!prob.empty()) {
            std::map<uint32_t,float>::iterator it=prob.begin();
            bestLabelIdx=it->first;
            maxProb=it->second;
            for (++it; it!=prob.end(); ++it)
               if (it->second>maxProb) {
                  maxProb=it->second;
                  bestLabelIdx=it->first;
               }
         }
         return bestLabelIdx;
      }

      uint32_t supportSize() const {
         int supportSize=0;
         for (std::map<uint32_t,float>::const_iterator it=prob.begin(); it!=prob.end(); ++it) {
            supportSize+=it->second>0;
         }
         return supportSize;
      }

      std::map<uint32_t,float> prob;

};

struct ShapeSample {
      uint32_t idx;
};

struct ShapeLabel {
      uint32_t value;
};

/**
 *
 */
class Shape
{
   private:
      std::vector<double> evecs;
      std::vector< std::pair<double,const double*> > eigenpairs;

      // (wks component idx , max frequency idx , frequency idx)
      std::vector< std::vector< std::vector<double> > > wks_gaussians;

   public:
      std::vector<int> labels;

      // we might want to consider only a subsampling of the shape.
      // leave empty to use all the points in the shape.
      std::vector<int> samples;
      bool use_all_points;

      Shape(const std::vector<int>& labels_, const std::vector<int>& samples_, const std::vector<double>& evals_, const std::vector<double>& evecs_, int max_nev, int descr_size);
      double compute_wks(int point_idx, int n_eigenvalues, int energy_level) const;

      inline double getHarmonic(int k, int p) const
      {
         return eigenpairs.at(k).second[p];
      }
};

/**
 *
 */
class DataSet
{
   private:

      std::vector<Shape*> shapes;
      std::vector<int> n_labels;
      size_t n_shapes;
      size_t n_tot_labels;
      size_t n_eigenvectors;
	  size_t n_energy_levels;

      void getShapeSample(uint32_t tidx, uint32_t& shapeIdx, uint32_t& sampleIdx) const;
      double getHarmonicValue(uint32_t shapeIdx, uint32_t sampleIdx, uint32_t harmonicIdx) const;

   public:

      explicit DataSet(size_t nev, size_t dim) : n_shapes(0), n_tot_labels(0), n_eigenvectors(nev), n_energy_levels(dim) {}

      virtual ~DataSet() throw()
      {
         clear();
      }

      inline void clear()
      {
         for (size_t i=0; i<n_shapes; ++i)
            delete shapes[i];
         shapes.clear();
         n_shapes = 0;
         n_labels.clear();
         n_tot_labels = 0;
      }

      inline void addShape(const std::vector<int>& labels, const std::vector<int>& samples, const std::vector<double>& evals, const std::vector<double>& evecs)
      {
         if (getWDPUpperBounds().nEigenVectors > evals.size())
            throw std::out_of_range("DataSet::addShape(): the training process is asking for more WKS frequencies than available.");

         Shape* shape = new Shape(labels, samples, evals, evecs, getWDPUpperBounds().nEigenVectors, getWDPUpperBounds().energy_level);
         shapes.push_back(shape);
         ++n_shapes;

         if (shape->use_all_points)
         {
            n_tot_labels += labels.size();
            n_labels.push_back(labels.size());
         }
         else
         {
            n_tot_labels += samples.size();
            n_labels.push_back(samples.size());
         }
      }

      inline double evaluateWaveDescriptor(uint32_t tsample, const WaveDescriptorParams &params) const
      {
         uint32_t shapeIdx, sampleIdx;
         getShapeSample(tsample, shapeIdx, sampleIdx);

         const Shape& shape = *shapes[shapeIdx];
         const int pointIdx = (shape.use_all_points ? sampleIdx : shape.samples[sampleIdx]);

         return shape.compute_wks(pointIdx , params.nEigenVectors, params.energy_level );
      }

      inline uint32_t getNumberOfSamples() const
      {
         return n_tot_labels;
      }

      inline uint32_t getLabel(uint32_t tsample) const
      {
         uint32_t shapeIdx, sampleIdx;
         getShapeSample(tsample, shapeIdx, sampleIdx);

         const Shape& shape = *shapes[shapeIdx];
         const int pointIdx = (shape.use_all_points ? sampleIdx : shape.samples[sampleIdx]);

         return shape.labels[pointIdx];
      }

      inline WaveDescriptorParams getWDPLowerBounds() const
      {
         WaveDescriptorParams params;
         params.energy_level = 0;
         params.nEigenVectors = 10;
         return params;
      }

      inline WaveDescriptorParams getWDPUpperBounds() const
      {
         WaveDescriptorParams params;
         //params.energy_level = 99;
		 params.energy_level = n_energy_levels-1;
         params.nEigenVectors = n_eigenvectors;
         return params;
      }
};

class ShapeRandomTree : public
      RandomTree<ShapeSplitData, ShapeSample, ShapeLabel, ShapePrediction> {
   public:
      ShapeRandomTree() :
         maxDepth(0), minSize(0), nTrials(1), dataset(NULL) {
         srand(time(NULL));
      }
	  
	  uint32_t getNumberOfLabels() const {
		 return dataset ? dataset->getNumberOfSamples() : 0;
	  }

      void setMaxDepth(uint32_t maxDepth) {
         this->maxDepth=maxDepth;
      }

      void setMinSize(uint32_t minSize) {
         this->minSize=minSize;
      }

      void setNTrials(uint32_t nTrials) {
         this->nTrials=nTrials;
      }
      void setDataSet(const DataSet *dataset) {
         this->dataset = dataset;
      }

      void train() {
         LSamplesVector samples;
         samples.resize(dataset->getNumberOfSamples());
         for (size_t i=0; i<samples.size(); ++i) {
            samples[i].sample.idx=i;
            samples[i].label.value=dataset->getLabel(i);
         }

         RandomTree<ShapeSplitData, ShapeSample, ShapeLabel, ShapePrediction>::train(samples);
      }

      ShapePrediction predict(int idx) const {
         ShapeSample s;
         s.idx=idx;
         return RandomTree<ShapeSplitData, ShapeSample, ShapeLabel, ShapePrediction>::predict(s);
      }

      bool split(const ShapeSplitData &split, const ShapeSample &sample) const {
         if (split.type==0)
            return dataset->evaluateWaveDescriptor(sample.idx, split.measurements[0])>split.thres;
         else
            return (dataset->evaluateWaveDescriptor(sample.idx, split.measurements[0])
                  -dataset->evaluateWaveDescriptor(sample.idx, split.measurements[1]))>split.thres;
      }

      void initialize(const RandomTreeNode *node, ShapePrediction &prediction) const {
         assert(node->getNSamples()>0);
         const LSamplesVector &samples=getLSamples();

         prediction.reset();
         for (int i=node->getStart(); i<node->getEnd(); ++i)
            prediction.add(samples[i].label.value, 1);
         prediction.normalize();
      }

      bool split(const RandomTreeNode *node, ShapeSplitData &splitData,
                 ShapePrediction &leftPrediction, ShapePrediction &rightPrediction) {

         //cvlab::HiresTimer timer; timer.start();

         bool found=false;

         if (node->getPrediction().prob.size()>1 && (maxDepth==0 || node->getDepth()<maxDepth)) {
            double bestLoss=numeric_limits<double>::infinity();
            ShapeSplitData newSplit;

            const LSamplesVector &samples=getLSamples();
            uint32_t nLeft=0, nRight=0;

            for (size_t t=0; t<nTrials; ++t) {

               generateSplit(node, newSplit);

               leftPrediction.reset();
               rightPrediction.reset();

               nLeft=0, nRight=0;
               for (int i=node->getStart(); i<node->getEnd(); ++i)
                  if (split(newSplit, samples[i].sample)) {
                     rightPrediction.add(samples[i].label.value, 1);
                     ++nRight;
                  } else {
                     leftPrediction.add(samples[i].label.value, 1);
                     ++nLeft;
                  }
               double newLoss=0;

               std::map<uint32_t,float>::iterator it;
               for (it=leftPrediction.prob.begin(); it !=leftPrediction.prob.end(); ++it) {
                  if (it->second>0)
                     newLoss-=it->second *log(it->second/nLeft);
               }

               for (it= rightPrediction.prob.begin(); it !=rightPrediction.prob.end(); ++it) {
                  if (it->second>0)
                     newLoss-=it->second *log(it->second/nRight);
               }
               if (newLoss<bestLoss && nLeft>minSize && nRight>minSize) {
                  splitData=newSplit;
                  bestLoss=newLoss;
                  found=true;
               }
            }

            if (found) { //recompute the best split data
               SplitResultsVector &results=getSplitResults();

               leftPrediction.reset();
               rightPrediction.reset();

               for (int i=node->getStart(); i<node->getEnd(); ++i) {
                  results[i]=split(splitData, samples[i].sample);
                  if (results[i])
                     rightPrediction.add(samples[i].label.value, 1);
                  else
                     leftPrediction.add(samples[i].label.value, 1);
               }
               leftPrediction.normalize();
               rightPrediction.normalize();

            }
         }

         //timer.stop();
         //std::cout << "Elapsed: " << timer.elapsed() << " seconds." << std::endl;

         return found;
      }

      void writeHeader(ostream &out) const {
      }

      void writeHeader_binary(ostream &out) const {
      }

      void readHeader(istream &in) {
      }

      void readHeader_binary(istream &in) {
      }

      void write(const ShapeSample &sample, ostream &out) const {
         out << sample.idx;
      }

      void write_binary(const ShapeSample &sample, ostream &out) const {
         boost::uint32_t sampleidx = sample.idx;
         out.write((const char*)&sampleidx,4);
      }

      void read(ShapeSample &sample, istream &in) const {
         in >> sample.idx;
      }

      void read_binary(ShapeSample &sample, istream &in) const {
         boost::uint32_t sampleidx;
         in.read((char*)&sampleidx,4);
         sample.idx = sampleidx;
      }

      void write(const ShapePrediction &prediction, ostream &out) const {
         out << prediction.prob.size();

         for (std::map<uint32_t,float>::const_iterator it=prediction.prob.begin(); it
              !=prediction.prob.end(); ++it)
            out << " " << it->first << " " << it->second;
      }

      void write_binary(const ShapePrediction &prediction, ostream &out) const {
         boost::uint32_t sz = (boost::uint32_t)prediction.prob.size();
         out.write((const char*)&sz,4);

         for (std::map<uint32_t,float>::const_iterator it=prediction.prob.begin(); it
              !=prediction.prob.end(); ++it)
         {
            boost::uint32_t first = it->first;
            float second = it->second;
            out.write((const char*)&first,4);
            out.write((const char*)&second,sizeof(float));
         }
      }

      void read(ShapePrediction &prediction, istream &in) const {
         prediction.reset();
         uint32_t n;
         in >> n;
         for (uint32_t i=0; i<n; ++i) {
            uint32_t idx;
            float val;
            in >> idx >> val;
            prediction.prob[idx]=val;
         }
      }

      void read_binary(ShapePrediction &prediction, istream &in) const {
         prediction.reset();
         boost::uint32_t n;
         in.read((char*)&n,4);
         for (boost::uint32_t i=0; i<n; ++i) {
            boost::uint32_t idx;
            float val;
            in.read((char*)&idx,4);
            in.read((char*)&val,sizeof(float));
            prediction.prob[idx]=val;
         }
      }

      void write(const ShapeLabel &label, ostream &out) const {
         out << label.value;
      }

      void write_binary(const ShapeLabel &label, ostream &out) const {
         boost::uint32_t val = label.value;
         out.write((const char*)&val,4);
      }

      void read(ShapeLabel &label, istream &in) const {
         in >> label.value;
      }

      void read_binary(ShapeLabel &label, istream &in) const {
         boost::uint32_t val;
         in.read((char*)&val,4);
         label.value = val;
      }

      void write(const ShapeSplitData &splitData, ostream &out) const {
         out << splitData.type << " " << splitData.thres<< " ";
         write(splitData.measurements[0], out);
         out << " ";
         write(splitData.measurements[1], out);
      }

      void write_binary(const ShapeSplitData &splitData, ostream &out) const {
         boost::int16_t type = splitData.type;
         double thresh = splitData.thres;
         out.write((const char*)&type,2);
         out.write((const char*)&thresh,sizeof(double));
         write_binary(splitData.measurements[0], out);
         write_binary(splitData.measurements[1], out);
      }

      void read(ShapeSplitData &splitData, istream &in) const {
         in >> splitData.type;
         in >> splitData.thres;
         read(splitData.measurements[0], in);
         read(splitData.measurements[1], in);
      }

      void read_binary(ShapeSplitData &splitData, istream &in) const {
         boost::int16_t type;
         double thresh;
         in.read((char*)&type,2);
         in.read((char*)&thresh,sizeof(double));
         splitData.type = type;
         splitData.thres = thresh;
         read_binary(splitData.measurements[0], in);
         read_binary(splitData.measurements[1], in);
      }

   private:

      void write(const WaveDescriptorParams &d, ostream &out) const {
         //out << d.energy_level << " " << d.variance << " " << d.nEigenVectors;
         out << d.energy_level << " " << d.nEigenVectors;
      }

      void write_binary(const WaveDescriptorParams &d, ostream &out) const {
         boost::int32_t e = d.energy_level;
         //double v = d.variance;
         boost::uint32_t ev = d.nEigenVectors;
         out.write((const char*)&e,4);
         //out.write((const char*)&v,sizeof(double));
         out.write((const char*)&ev,4);
      }

      void read(WaveDescriptorParams &d, istream &in) const {
         //in >> d.energy_level >> d.variance >> d.nEigenVectors;
         in >> d.energy_level >> d.nEigenVectors;
      }

      void read_binary(WaveDescriptorParams &d, istream &in) const {
         boost::int32_t e;
         //double v;
         boost::uint32_t ev;
         in.read((char*)&e,4);
         //in.read((char*)&v,sizeof(double));
         in.read((char*)&ev,4);
         d.energy_level = e;
         //d.variance = v;
         d.nEigenVectors = ev;
      }

      void generateSplit(const RandomTreeNode *node, ShapeSplitData &split) const {

         const LSamplesVector &samples=getLSamples();
         split.type=rand()%2;
         WaveDescriptorParams ub=dataset->getWDPUpperBounds();
         WaveDescriptorParams lb=dataset->getWDPLowerBounds();
         WaveDescriptorParams p;
         p.nEigenVectors=lb.nEigenVectors+rand()%(ub.nEigenVectors-lb.nEigenVectors);
         p.energy_level=lb.energy_level+rand()%(ub.energy_level-lb.energy_level);
         //p.variance=lb.variance+rand()*(ub.variance-lb.variance)/RAND_MAX;
         split.measurements[0]= p;
         p.nEigenVectors=lb.nEigenVectors+rand()%(ub.nEigenVectors-lb.nEigenVectors);
         p.energy_level=lb.energy_level+rand()%(ub.energy_level-lb.energy_level);
         //p.variance=lb.variance+rand()*(ub.variance-lb.variance)/RAND_MAX;
         split.measurements[1]= p;

         double min_range=numeric_limits<double>::infinity();
         double max_range=-min_range;
         double val;

         for (int i=node->getStart(); i<node->getEnd(); ++i) {
            if (split.type==0)
               val=dataset->evaluateWaveDescriptor(samples[i].sample.idx, split.measurements[0]);
            else
               val=dataset->evaluateWaveDescriptor(samples[i].sample.idx, split.measurements[0])
                     -dataset->evaluateWaveDescriptor(samples[i].sample.idx, split.measurements[1]);
            if (min_range>val)
               min_range=val;
            if (max_range<val)
               max_range=val;
         }
         split.thres=min_range+(max_range-min_range)*((double)rand() /RAND_MAX);
      }

      uint32_t maxDepth;
      uint32_t minSize;
      uint32_t nTrials;

      const DataSet *dataset;
};

static ShapePrediction forest_prediction(const vector<ShapeRandomTree> &forest, int idx)
{
   ShapePrediction finalPrediction=forest[0].predict(idx);
   for (size_t i=1; i<forest.size(); ++i)
      finalPrediction.add(forest[i].predict(idx));
   finalPrediction.normalize();
   return finalPrediction;
}

}

#endif /*SIMPLERFCLASSIFIER_H_*/
