/**
 * This code accompanies the paper:
 *
 * "Dense non-rigid shape correspondence using random forests"
 * E.Rodola, S.Rota Bulo, T.Windheuser, M.Vestner, D.Cremers.
 * Proc. CVPR 2014
 *
 * Please cite the paper above if you use the code.
 */
 
#ifndef MESH_H
#define MESH_H

#include <vector>
#include <list>
#include <string>

template <typename T>
struct vec3d
{
  T x,y,z;

  explicit vec3d() : x(0), y(0), z(0) {}
  vec3d(T x_, T y_, T z_) : x(x_), y(y_), z(z_) {}

  template <typename S>
  vec3d(const vec3d<S>& s) : x(T(s.x)), y(T(s.y)), z(T(s.z)) {}
};

class mesh_t
{
private:
	
   struct vertex_data {
      vertex_data() : n_tris(0) {}
      vec3d<double> p;
      size_t n_tris; // no. of triangles this vertex belongs to
      std::list<int> tri_indices; // triangles this vertex belongs to
   };

   struct triangle_data {
      int p0, p1, p2; // indices of the 3 vertices
   };
   
   std::vector<vertex_data> vertices;
   std::vector<triangle_data> triangles;
   
public:
   
   inline size_t mesh_t::size() const
   {
		return vertices.size();
   }
   
   void mesh_t::put_vertices(const std::vector< vec3d<double> >& points)
	{
	   const size_t n_points = points.size();
	   vertices.resize(n_points);

	   for (size_t k=0; k<n_points; ++k)
	   {
		  // indexed vertices
		  vertex_data v;
		  v.p = vec3d<double>(points[k]);
		  v.n_tris = 0;
		  vertices[k] = v;
	   }
	}
	
	void mesh_t::add_triangle(int p0, int p1, int p2)
	{
	   triangle_data td;

	   td.p0 = p0;
	   td.p1 = p1;
	   td.p2 = p2;

	   triangles.push_back(td);
	   const int tri_idx = static_cast<int>(triangles.size() - 1);

	   vertices.at(p0).tri_indices.push_back(tri_idx);
	   vertices.at(p0).n_tris++;
	   vertices.at(p1).tri_indices.push_back(tri_idx);
	   vertices.at(p1).n_tris++;
	   vertices.at(p2).tri_indices.push_back(tri_idx);
	   vertices.at(p2).n_tris++;
	}
};
   
#endif
